;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: boolprop.lsp
;;;;    System: LTRE
;;;;   Version: 4.1 
;;;;    Author: Ken Forbus, Johan de Kleer, John Everett
;;;;   Created: May 1, 1986
;;;;  Modified: everett on Mon Jul 1 22:37:45 1996
;;;;   Purpose: Boolean constraint propagation
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)


;;; BOOLEAN CONSTRAINT PROPAGATION
;;; ----------------------------------------------------------------------------
;;; *Clauses-to-check* is a pointer to the thread of clauses that the ltre
;;; maintains during retraction operations.

(defvar *clauses-to-check* nil "Pointer to clause thread")

(defun check-clauses (ltms *clauses-to-check*)
   (dbg-ltre :clauses "Checking ~A clauses..." (clause-queue-length))
   (setf (ltms-clause-thread-dirty? ltms) T)
   (do () ((null *clauses-to-check*))
      (check-clause ltms (pop-clauses-to-check-thread)))
   (setf (ltms-clause-thread-dirty? ltms) nil))

(defun check-clause (ltms clause &aux unknown-pair)
   (dbg-ltre :clauses "Checking: ~A" clause)
   (cond ((violated-clause? clause)
          ;;pushnew instead of push for safety across interrupts
          (pushnew clause (ltms-violated-clauses ltms)))
         ((= (clause-pvs clause) 1)
          ;;Exactly one term of the clause remains that can
          ;;satisfy the clause, so deduce that term
          (setq unknown-pair (find-unknown-pair clause))
          (when unknown-pair ;;Must check for alternative support
             (set-truth (car unknown-pair) 
              (cdr unknown-pair) clause)))))

(defun find-unknown-pair (clause)
  (dolist (term-pair (clause-literals clause))
    (if (unknown-node? (car term-pair)) (return term-pair))))

(defun top-set-truth (node value reason)
   (declare (special *clauses-to-check*))
   (let ((*clauses-to-check* nil))
      (set-truth node value reason)
      (check-clauses (tms-node-ltms node) *clauses-to-check*)
      (check-for-contradictions (tms-node-ltms node))))

(defun set-truth (node value reason)
   (setf (tms-node-support node) reason)
   (setf (tms-node-label node) value)
   (ecase value ;;figure out which set of rules to queue up
      (:TRUE 
       (enqueue-cached-true-rules node)
       (dolist (clause (tms-node-true-clauses node))
          (incf (clause-sats clause)))
       (dolist (clause (tms-node-false-clauses node))
          (when (< (decf (clause-pvs clause)) 2)
             (add-to-clauses-to-check-thread clause)))) 
      (:FALSE 
       (enqueue-cached-false-rules node)
       (dolist (clause (tms-node-false-clauses node))
          (incf (clause-sats clause)))
       (dolist (clause (tms-node-true-clauses node))
          (when (< (decf (clause-pvs clause)) 2)
             (add-to-clauses-to-check-thread clause))))))


;;; THREADING CLAUSES
;;; ----------------------------------------------------------------------------
;;; To minimize consing, we thread all clauses within set-truth.  If 
;;; *clauses-to-check* is nil, then it is pointed at the first threaded 
;;; clause.  Thereafter, that clause's clause-thread pointer is pointed at the 
;;; next clause to be threaded.  By chasing these thread pointers, we can 
;;; recover the clauses to check queue.  Have to be very careful not to create 
;;; circular threads, because such structures can cause ungraceful crashes.  The
;;; clause-thread is a push-down stack in which *clauses-to-check* always points
;;; to the last clause added.

(defun add-to-clauses-to-check-thread (clause)
   "Return top of a virtual queue of clauses"
   (cond ((or (eq *clauses-to-check* clause)
              (clause-thread clause))
          nil) ;;no circular threads, please!
         ((null *clauses-to-check*)
          (setq *clauses-to-check* clause))
         ((and (clause-p *clauses-to-check*)
               (null (clause-thread clause)))
          (setf (clause-thread clause) *clauses-to-check*)
          (setq *clauses-to-check* clause))
         (t (error "Clause thread weirdness: Clause: ~A~%Clauses-to-check: ~A~%"
               clause *clauses-to-check*))))
   
(defun pop-clauses-to-check-thread ()
   (let ((popped-clause *clauses-to-check*))
      ;;note that (shiftf a b c) ==> (setq a b) (setq b c)
      (shiftf *clauses-to-check* (clause-thread popped-clause) nil)
      popped-clause))

(defun thread-true-clauses (node)
   "Add true-clauses of node to the *clauses-to-check* virtual queue"
   (dolist (clause (tms-node-true-clauses node))
      (add-to-clauses-to-check-thread clause))
   *clauses-to-check*)

(defun thread-false-clauses (node)
   "Add false-clauses of node to the *clauses-to-check* virtual queue"
   (dolist (clause (tms-node-false-clauses node))
      (add-to-clauses-to-check-thread clause))
   *clauses-to-check*)

(defun show-clause-queue (&aux (indices nil))
   "Show clause thread for debugging"
   (do* ((clause *clauses-to-check* (clause-thread clause)))
        ((null clause) 
         (format *trace-output* "~&~A~%" (nreverse indices)))
      (push clause indices)))

(defun clause-queue-length (&aux (count 0))
   (do* ((clause *clauses-to-check* (clause-thread clause)))
        ((null clause) count)
      (incf count)))

(defun show-node-thread (head &aux result)
   (do* ((node head (tms-node-thread node)))
        ((null node) 
         (format *trace-output* "~&Node Thread: ~A~%"
            (nreverse result)))
      (push node result)))


;;; RETRACTING AN ASSUMPTION
;;; ----------------------------------------------------------------------------
;;; Propagate-unknownness threads nodes as it propagates to avoid consing.
;;; 1. Retain a pointer (head) to node passed in and return this node at end
;;; 2. Set up two fingers, current and tail, both at start pointing to in-node
;;; 3. Get clauses corresponding to node's label (either true or false)
;;; 4. Set the current node's label to :UNKNOWN
;;; 5. For each clause of node, increment its potential violators.
;;; 6. When this increment results in two pvs, set the thread pointer of the 
;;;    tail of the node thread to point to the consequent of the clause.  This 
;;;    consequent will either be another node or nil.  If it's a threaded node
;;;    then we have a loop in the TMS and the whole database is corrupt.  The 
;;;    error check has never caught a loop, and so is commented out.
;;; 7. When the tail's node-thread is a node, move the tail finger to point to 
;;;    the new node that has just been added to the thread.
;;; 8. After steps 5-7 are done for all clauses of node, there will be zero or 
;;;    more new nodes added to thread, each of which must be made unknown and 
;;;    must have its clauses checked.  Therefore we increment the current 
;;;    pointer, which now points to an unknown node, to point to the next node 
;;;    on the thread.
;;; 9. New nodes can therefore pile up on the end of the thread, but after a 
;;;    no new nodes will be added, and since the current finger steadily 
;;;    advances down the thread one node for each iteration of the do loop, it 
;;;    will eventually reach the end of the thread, which points to nil, so we 
;;;    use (null current) check in the do loop for termination.

(defun propagate-unknownness (in-node)
   (let ((head in-node) (current in-node) (tail in-node))
      (dbg-ltre :make-unknown "~&Propagating unknownness for ~A~%" in-node)
      (do ((clauses (clauses-for-propagation current)
                    (clauses-for-propagation current)))
          ((null current))
         (dbg-ltre :make-unknown "~&Making ~A unknown~%" current)
         (make-unknown current)
         (dolist (clause clauses)
            (when (= (incf (clause-pvs clause)) 2)
               (setf (tms-node-thread tail)
                     (clause-consequent clause)) ;;could be nil, so 
               (when (tms-node-thread tail)      ;;move tail only when not nil
                  ;; (check-for-node-loops (tms-node-thread tail)) ;;debugging
                  (setq tail (tms-node-thread tail))
                  (dbg-ltre :make-unknown "~%Setting tail pointer to ~A" tail))))
         (setq current (tms-node-thread current))
         (dbg-ltre :make-unknown "~%Setting current pointer to ~A" current))
      head))

(defun clauses-for-propagation (in-node)
   (when (tms-node-p in-node)
      (ecase (tms-node-label in-node)
         (:TRUE (tms-node-false-clauses in-node))
         (:FALSE (tms-node-true-clauses in-node)))))

(defun make-unknown (node)
   (setf (tms-node-label node) :UNKNOWN)
   (setf (tms-node-support node) NIL))
 
(defun clause-consequent (clause)
   (dolist (term-pair (clause-literals clause))
      (when (eq (tms-node-label (car term-pair)) (cdr term-pair))
         (return (if (eq clause (tms-node-support (car term-pair)))
                    (car term-pair))))))

(defun check-for-node-loops (node)
   "If new node has a thread pointer to another node, then all bets are off"
   (when (tms-node-thread node)
      (error "~A is already threaded, database integrity suspect" node)))

(defun find-alternative-support (ltms node)
   "Scan node thread, checking clauses of each unknown node"
   (do ((next-node node (tms-node-thread next-node)))
       ((null next-node) node)
      (when (unknown-node? next-node)
         (dbg-ltre :alt-support "Seeking alternative support for: ~A tc: ~A fc: ~A"
          next-node (tms-node-true-clauses next-node)
          (tms-node-false-clauses next-node))
         ;;Must start with clean thread here to avoid doing off-topic work
         (let ((*clauses-to-check* nil))  
            (check-clauses ltms (thread-true-clauses next-node)))
         (let ((*clauses-to-check* nil))
            (check-clauses ltms (thread-false-clauses next-node))))))

;;; ----------------------------------------------------------------------------
;;; END OF CODE