;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: cwa.lsp
;;;;    System: LTRE
;;;;   Version: 4.1
;;;;    Author: Kenneth D. Forbus & Johan de Kleer
;;;;   Created: 10/1/90
;;;;  Modified: everett on Sat Jul 6 13:47:18 1996
;;;;   Purpose: Closed world assumption management, new version as of 6/94
;;;; Copyright: (c) 1986-1993, Kenneth D. Forbus, Northwestern University,
;;;;            and Johan de Kleer, the Xerox Corporation.
;;;;            All rights reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)

;;; GENERAL
;;; ----------------------------------------------------------------------------
;;; RETRACT-CWA now retracts a CWA, inferred or assumed, so that the MEMBERS
;;; statement implied the the state of the current database can be justified.
;;; If this leads to a contradiction, so be it!  Metering indicates that the 
;;; fetches involved in this system are a major source of runtime consing, so 
;;; this version includes some special-case procedures that violate the LTRE 
;;; database abstraction, replacing FETCH with special-purpose routines, to
;;; speed things up.


;;; INTERFACE TO SET MECHANISM
;;; ----------------------------------------------------------------------------
(eval-when (eval compile load)
  (proclaim '(special *set-rule-file*)))

(defun set-members (set-name &optional (*LTRE* *LTRE*) &aux form)
   ;; Modularity violation for efficiency: Cost of UNIFY dominated in previous
   ;; version. Still assumes that only one MEMBERS statement holds at a time
   (dolist (datum (dbclass-facts (get-dbclass set-name)))
      ;; An open-coding of unify for (<set name> MEMBERS <members>)
      (setq form (datum-lisp-form datum))
      (when (and (equal (car form) set-name)
                 (equal (cadr form) 'cl-user::MEMBERS)
                 (null (cdddr form))
                 (true-node? (datum-tms-node datum)))
         (return-from SET-MEMBERS
            (values (third form) (find-cwa-for-set form))))))

(defun close-set-if-needed (set-name &optional (*LTRE* *LTRE*))
   (multiple-value-bind (construal cwa)
       (set-members set-name)
      (if cwa
         (values construal cwa nil)
         (close-set set-name))))

(defun close-set (set-name &optional (*LTRE* *LTRE*))
   (let ((non-contradictory? nil))
      (unwind-protect
           (multiple-value-bind (known-members known-not)
               (get-set-information set-name)
              (let ((cwa-form (make-cwa-form set-name known-members))
                    (members-form
                     `(,set-name CL-USER::MEMBERS ,known-members)))
                 (retract-CWAs set-name)
                 (assume-cwa-if-needed cwa-form)
                 (justify-cwa-if-needed set-name known-members known-not
                    cwa-form members-form)
                 (setq non-contradictory? t)
                 (values known-members cwa-form t)))
         (progn
            (unless non-contradictory?
               (retract-CWAs set-name))))))

(defmacro With-Closed-Set (set-name &rest body)
  `(multiple-value-bind (members cwa)
       (close-set-if-needed ,set-name *LTRE*)
     (With-Contradiction-Handler (ltre-ltms *LTRE*)
       #'(lambda (clauses ltms)
	   (set-cwa-handler clauses ltms ,set-name 
			    cwa :LOST-CWA))
       (let ((answer (catch cwa ,@body)))
	 (cond ((eq answer :LOST-CWA) (values nil nil))
	       (t (values t members)))))))


;;; 2ND-LEVEL PROCEDURES -- USE IF YOU KNOW WHAT YOU'RE DOING
;;; ----------------------------------------------------------------------------
(defun find-cwa-for-set (m-s)
  ;; This turns out to be very efficient
  (dolist (asn (assumptions-of m-s)) 
	  (when (and (cwa-form? asn)
		     (equal (car asn) (car m-s)))
		(return-from FIND-CWA-FOR-SET asn))))

(defun cwa-form? (form) (eq (cadr form) 'CWA))

(defun make-cwa-form (set-name members)
  `(,set-name CWA ,members))

(defun parse-cwa-form (cwa-form)
  (values (car cwa-form) (caddr cwa-form)))

(defun get-set-information (set-name &optional (*LTRE* *LTRE*))
   (let ((known-in nil) (known-out nil)(form nil))
      ;;Violate abstraction a bit for speed -- metering shows this is worth it.
      (dolist (possible (dbclass-facts (get-dbclass set-name *LTRE*)))
         (when (and (equal set-name (car (setq form (datum-lisp-form possible))))
                    (eq (cadr form) 'CL-USER::HAS-MEMBER))
            ;;Assumes that there is a 3rd element and no others.
            (cond ((true-node? (datum-tms-node possible))
                   (push (caddr form) known-in))
                  ((false-node? (datum-tms-node possible))
                   (push (caddr form) known-out)))))
      (values known-in known-out)))

(defun assume-cwa-if-needed (cwa-form)
   (when (false? cwa-form) ;; Going to see a contradiction soon!
      (propagate-unknownness (get-tms-node cwa-form)))
   (assume! cwa-form :CWA))

(defun justify-cwa-if-needed (name members not-members
                              cwa-form members-form)
   (let ((cwa (referent cwa-form t)))
      (unless (assoc (list members not-members)
                 (getf (datum-plist cwa) :CWA-JUSTS)
                 :TEST 'equal)
         (push (list members not-members)
            (getf (datum-plist cwa) :CWA-JUSTS))
         (assert!
          `(CWA-JUSTIFICATION
            (:AND (CL-USER::SET ,name)
             ,@(mapcar #'(lambda (el)
                            `(,name CL-USER::HAS-MEMBER ,el))
                  members)
             ,@(mapcar #'(lambda (el)
                            `(:NOT (,name CL-USER::HAS-MEMBER ,el)))
                  not-members)
             ,cwa-form)
            ,members-form)
          :SET-CWA-CLOSURE))))

(defun fetch-cwa-for (name)
   (dolist (cwa (fetch `(,name CWA . ?x)))
      (when (true? cwa) (return-from fetch-cwa-for cwa))))

(defun retract-cwa-for (name)
  (let ((cwa (fetch-cwa-for name)))
    (when cwa  ;; safety
	  (retract! cwa (informant-of cwa)))))


;;; CONTRADICTION HANDLING AND BACKING UP FOR CWAs
;;; ----------------------------------------------------------------------------
;;; This procedure detects when a CWA lurks within one of the LTMS's
;;; contradictions, and does a THROW to a designated tag to signal the
;;; assumer that the set is no longer believed. 

(defun set-cwa-handler (clauses ltms set-name cwa tag &aux asns losers)
   (let ((cwa-node (get-tms-node cwa)))
      (dolist (cl clauses) ;; For each contradictory clause,
         (setq asns (assumptions-of-clause cl))
         (when (and (member cwa-node asns) ;;Quick filter
                    (CWA-invalid? cwa))
            (retract-CWA cwa)
            (throw cwa tag)))))

(defun retract-CWA (cwa)
   (retract! cwa (informant-of cwa)))

(defun CWA-invalid? (cwa)
  (multiple-value-bind (set-name presumed-els)
   (parse-cwa-form cwa)
   (dolist (el presumed-els)
    (unless (true? `(,set-name CL-USER::HAS-MEMBER ,el)) 
	    (return-from CWA-invalid? T)))
   (dolist (hm-form (fetch `(,set-name cl-user::has-member ?el)))
    (when (true? hm-form)
     (unless (member (third hm-form) presumed-els
		     :TEST #'equal)
	     (return-from CWA-invalid? t))))))

(defun retract-CWAs (set &optional (*LTRE* *LTRE*) &aux form node)
 ;; Violate FETCH abstraction here because metering indicates it's significant
  (dolist (possible-cwa (dbclass-facts (get-dbclass set)))
    (setq form (datum-lisp-form possible-cwa))
    (when (and (listp form) ;; (<set> CWA <members>) is the form
	       (equal (car form) set)
	       (eq 'CWA (cadr form))
	       (null (cdddr form)) ;; End of form tests
	       (known-node?
		(setq node (datum-tms-node possible-cwa)))) ;; and known
	  ;; Then retract it.  Also open-code the appropriate part of retract!
	  ;; here, since we don't want to keep going through the database
	  ;; multiple times.
	  (retract-assumption node))))


;;; DEBUGGING CWAs
;;; ----------------------------------------------------------------------------
(defun show-CWAs-for (set)
  (dolist (m (fetch `(,set CWA ?members)))
   (why? m)))

(defun show-set-information (set)
  (dolist (m (fetch `(,set cl-user::has-member ?m)))
    (why? m))
  (dolist (m (fetch `(,set cl-user::members ?m)))
    (why? m))
  (dolist (m (fetch `(,set CWA ?m)))
	  (why? m)))


;;; A SIMPLE EXAMPLE
;;; ----------------------------------------------------------------------------
(defun cwa-interactive-test (&optional (debugging? nil))
   (In-LTRE (create-ltre "CWA Test" :DEBUGGING debugging?))
   (load-file *ltre-path* *set-rule-file*)
   (dolist (data '((cl-user::set (Parts System))  ;; Assume initial parts
                   ((Parts System) CL-USER::HAS-MEMBER valve)
                   ((Parts System) CL-USER::HAS-MEMBER meter)
                   ((Parts System) CL-USER::HAS-MEMBER pump)))
      (assume! data :INITIAL-OBSERVATIONS))
   (do ((form nil)
        (stop? nil)
        (partslist nil))
       (stop? nil)
      (With-Closed-Set
       '(Parts System)
       (setq partslist 
          (remove-if-not #'(lambda (form) (true? form))
             (fetch `((Parts System) CL-USER::MEMBERS ?els))))
       (cond ((cdr partslist)
              (format t "~%BUG: Conflicting membership forms.")
              (dolist (pl partslist)
                 (format t "Parts(System) = ~A" pl)))
             (t (format t "~% Parts are: ~A" (third (car partslist)))))
       (cond ((member form '(quit stop end exit)) (setq stop? t))
             (t (format t "~%>")
                (setq form (read))
                (print (eval form))
                (run-rules))))))


;;; SHAKEDOWN PROCEDURE
;;; ----------------------------------------------------------------------------
(defun cwa-shakedown ()
   (in-ltre (create-ltre "CWA Test One" :dbg-flags t))
   (load-file *ltre-path* *set-rule-file* :action :load-source)
   (uassert! '(cl-user::set foo) :USER)
   (with-closed-set 'foo
      (uassume! '(foo cl-user::has-member a) :A-IN))
  (with-closed-set 'foo
		   (uassume! '(foo cl-user::has-member b) :B-IN))
  (with-closed-set 'foo (run-rules))
  (if (multiple-value-bind (members cwa) (set-members 'foo)
			   ;; Must be same elements
       (and (equal (intersection members '(A B)) '(A B))
	    (equal (union members '(A B)) '(A B))))
      (format t "~% (A B) closure okay.")
     (break "1. Set FOO with members A B didn't close correctly"))
  (with-closed-set 'foo
		   (uassume! '(foo cl-user::has-member c) :C-IN))
  (with-closed-set 'foo (run-rules))
  (if (multiple-value-bind (members cwa) (set-members 'foo)
       (and (equal (intersection members '(A B C)) '(A B C))
	    (equal (union members '(A B C)) '(A B C))))
      (format t "~% (A B C) closure okay.")
    (break "2. C not properly added to set Foo"))
  (retract! '(foo cl-user::has-member a) :A-IN *ltre* nil)
  (with-closed-set 'foo (run-rules))
  (if (multiple-value-bind (members cwa) (set-members 'foo)
       (and (equal (intersection members '(B C)) '(B C))
	    (equal (union members '(B C)) '(B C))))
      (format t "~% Retraction handled okay.")
    (break "3. Removal of A from set Foo botched"))
  (retract-CWAs 'foo)
  (uassume! '(foo cl-user::has-member a) :A-IN)
  (with-closed-set 'foo (run-rules))
  (if (multiple-value-bind (members cwa) (set-members 'foo)
       (and (equal (intersection members '(A B C)) '(A B C))
	    (equal (union members '(A B C)) '(A B C))))
      (format t "~% Unouting handled okay.")
    (break "4. Unouting A botched"))
  (format t "~% If no breaks occurred, CWA's look okay."))

;;; ----------------------------------------------------------------------------
;;; END OF CODE  