;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: dds.lsp
;;;;    System: LTRE
;;;;   Version: 4.1
;;;;    Author: Kenneth D. Forbus & Johan de Kleer
;;;;   Created: 10/1/90
;;;;  Modified: everett on Sat Jul 6 13:49:38 1996
;;;;   Purpose: Dependency-directed search facility
;;;; Copyright (c) 1986-1993, Kenneth D. Forbus, Northwestern University,
;;;;            and Johan de Kleer, the Xerox Corporation.
;;;;            All rights reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)

;;; GLOBALS AND GENERAL
;;; ----------------------------------------------------------------------------
(defvar *debug-dds* nil)
(defvar *debug-dds-contradiction-processing* nil)

(defmacro When-DDS-Contradiction-debugging  (&rest forms)
   `(when *debug-dds-contradiction-processing* 
       ,@forms))

(defmacro When-DDS-Debugging (&rest forms)
   `(when *debug-dds* ,@forms))

(defmacro debug-dds (str &rest args)
   `(if *debug-dds* (format t ,str ,@args)))


;;; DD-SEARCH FUNCTIONS
;;; ----------------------------------------------------------------------------
(defun DD-Search (choice-sets end &aux answer marker choices)
   (when (null choice-sets)
      (debug-dds "~%    DDS: Found solution.")
      (eval end)
      (return-from DD-Search nil))
   (setq marker (list 'DDS (car choice-sets)))
   (setq choices (car choice-sets))
   (dolist (choice choices)
      (debug-dds "~%    DDS: Considering ~A..." choice)
      (cond ((false? choice) ;skip if known loser
             (debug-dds "~%    DDS: ~A already known nogood." choice))
            ((true? choice) ;skip useless if known
             (debug-dds "~%    DDS: ~A true by implication." choice)
             (DD-Search (cdr choice-sets) end)
             (return nil))
            (t (debug-dds "~%    DDS: Assuming ~A." choice)
               (with-Contradiction-Handler (ltre-ltms *ltre*)
                #'(lambda (clauses ltms &aux asns)
                     (declare (ignore LTMS))
                     (debug-dds "~%    DDS: Entering handler for ~A with ~A~A."
                      choice clauses
                      (mapcar #'(lambda (c) (violated-clause? c))
                         clauses)) 
                     (dolist (cl clauses)
                        (setq asns (assumptions-of-clause cl))
                        (debug-dds "~%    DDS: Assumptions are: ~A"
                         (mapcar #'view-node asns))
                        (dolist (asn asns)
                           (when (or (equal choice (view-node asn))
                                     (and (listp choice) (eq (car choice) :NOT)
                                          (equal (cadr choice) (view-node asn))))
                              (throw marker
                                 (cons :LOSERS
                                    ;; Assign labels before any retraction
                                    (mapcar #'signed-view-node
                                       (delete asn asns))))))))
                (setq answer (catch marker
                                (Assuming (list choice) *ltre*
                                   (run-rules *ltre*)
                                   (DD-Search (cdr choice-sets) end))))
                (when (and (listp answer)
                           (eq (car answer) :LOSERS))
                   (debug-dds "~%    DDS: ~A inconsistent with ~A."
                      choice (cdr answer))
                   (assert! `(:NOT (:AND ,choice
                                    ,@(cdr answer)))
                    :DD-SEARCH-NOGOOD)))))))


;;; Below is a version of DDS search that takes a procedure of no arguments
;;; to be executed in all consistent solution environments

(defun DD-Search-HOP (choice-sets proc &aux answer marker choices)
   ;; HOP = higher-order procedure.  PROC in this case is a procedure
   ;; of no arguments which is executed when a solution is found
   ;; to do whatever should be done in that environment.
   (when (null choice-sets)
      (debug-dds "~%    DDS: Found solution.")
      (funcall proc)
      (return-from DD-Search-HOP nil))
   (setq marker (list 'DDS (car choice-sets)))
   (setq choices (car choice-sets))
   ;;; ***** Install case-specific debugging hooks here
   ;;; ***** End of case-specific debugging hooks.
   (dolist (choice choices)
      (debug-dds "~%    DDS: Considering ~A..." choice)
      (cond ((false? choice) ;skip if known loser
             (debug-dds "~%    DDS: ~A already known nogood." choice))
            ((true? choice) ;skip useless if known
             (debug-dds "~%    DDS: ~A true by implication." choice)
             (DD-Search-HOP (cdr choice-sets) proc)
             (return nil))
            (t (debug-dds "~%    DDS: Assuming ~A." choice)
               (with-Contradiction-Handler (ltre-ltms *ltre*)
                #'(lambda (clauses ltms)
                     (DD-Search-Hop-Contradiction-Handler
                      choice marker clauses ltms))
                (setq answer (catch marker
                                (new-assuming (list choice) marker *ltre*
                                   (run-rules *ltre*)
                                   (DD-Search-HOP (cdr choice-sets) proc))))
                (when (and (listp answer)
                           (eq (car answer) :LOSERS))
                   (debug-dds "~%    DDS: ~A inconsistent with ~A."
                      choice (mapcar #'(lambda (node)
                                          (cons (view-node node)
                                             (tms-node-label node)))
                                (cdr answer)))
                   (unless (some #'unknown-node? (cdr answer))
                      (assert! `(:NOT (:AND ,choice
                                       ,@ (mapcar #'signed-view-node
                                             (cdr answer))))
                         :DD-SEARCH-NOGOOD))))))))

(defun DD-Search-HOP-Contradiction-Handler
  (choice marker clauses ltms &aux asns)
  (declare (ignore LTMS))
  (When-DDS-Contradiction-debugging
   (format t "~%    DDS: Entering handler for ~A with~%   ~A~%  ~A."
	   choice clauses
	   (mapcar #'(lambda (c) (violated-clause? c))
		   clauses)))
  (dolist (cl clauses)
   (setq asns (assumptions-of-clause cl))
   (dolist (asn asns)
	   (when (or (equal choice (view-node asn))
		     (and (listp choice) (eq (car choice) :NOT)
			  (equal (cadr choice) (view-node asn))))
		 (When-DDS-Contradiction-Debugging
		  (format t "~%    DDS: Handling it.  Assumptions are:")
		  (dolist (asn asns)
			  (format t "~%  ~A" (signed-view-node asn))))
		 (throw marker
			(cons :LOSERS (delete asn asns)))))))


;;; A FAMILIAR EXAMPLE
;;; ----------------------------------------------------------------------------
(defun Test-DD-search (&optional (debugging? t))
  (in-LTRE (create-ltre "DDS Test" :DEBUGGING debugging?))
  (eval '(rule ((:TRUE A) (:TRUE C))
	       (rassert! (:NOT (:AND A C)) :DOMAIN-NOGOOD)))
  (eval '(rule ((:TRUE B) (:TRUE E))
	       (rassert! (:NOT (:AND B E)) :DOMAIN-NOGOOD)))
  (DD-Search '((A B) (C D) (E F)) 
	     '(show-DD-test-solution)))
	    
(defun show-DD-test-solution (&aux result)
  (dolist (var '(F E D C B A))
    (when (true? var *ltre*) (push var result)))
  (format t "~% Consistent solution: (~A)." result))

;;; ----------------------------------------------------------------------------
;;; END OF CODE