;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: lcache.lsp
;;;;    System: LTRE
;;;;   Version: 4.1
;;;;    Author: Ken Forbus, Yusuf Pisan
;;;;   Created: Sept, 1994
;;;;  Modified: everett on Sat Jul 6 20:26:30 1996
;;;;   Purpose: Cache large sets of rules to avoid reloading them from disk
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)


;;; GENERAL
;;; ----------------------------------------------------------------------------
;;; Turns out that incorporating a debugging feature that saves rule body info 
;;; in rule structs is especially handy, so we define two versions of the 
;;; caching function, one for debugging, the other for production code.

;;; To use this caching  mechanism, define functions of the following form:
(defun load-cyclepad-kb ()
   (declare (special *cyclepad-kb-rules* *cyclepad-kb-facts*))
   (multiple-value-setq (*cyclepad-kb-rules* *cyclepad-kb-facts*)
      (make-ltre-cache '(load-cycle-rules-into-cache)))
   (dotimes (i 3)
      (beep *terminal-io*)
      (if (= i 2) (sleep 0.24) (sleep 0.1)))
   (values (length *cyclepad-kb-rules*) (length *cyclepad-kb-facts*)))

(defun install-cyclepad-kb ()
  (dolist (index-form *cyclepad-kb-rules*)
     (apply #'ltre::insert-rule
        (ltre::get-dbclass (car index-form))
        (cdr index-form)))
  (dolist (fact-form *cyclepad-kb-facts*) 
     (apply (car fact-form) (cdr fact-form))))


#-DEBUG-LTRE-RULES
(defun make-ltre-cache (&rest forms)
  ;; For each procedure that installs stuff
  ;; into an LTRE, redefine so that it instead
  ;; caches an appropriate calling form onto
  ;; one of rule-cache or fact-cache
  (let ((rule-cache nil)
        (fact-cache nil)
        (safe-insert-rule (symbol-function 'insert-rule))
        (safe-assert! (symbol-function 'assert!))
        (safe-assume! (symbol-function 'assume!)))
     ;; if an exit occurs in cleanup form we are in trouble
     (unwind-protect
          (progn
             (setf (symbol-function 'insert-rule)
                   ;;;Add var name after to lambda list after dbclass when debugging rules
                   #'(lambda (dbclass matcher body)
                        ;;(format t "~%Insert-rule called on: dbclass: ~S matcher: ~S body: ~S"
                        ;;   dbclass matcher body)
                        ;; ACLPC is fussy about using symbols as procedures
                        ;; In fact, we have to place function objects in cache
                        ;; in order to avoid compile problems in runtime
                        (push `(,(dbclass-name dbclass)
                                ,(if (symbolp matcher)
                                    `,(symbol-function matcher)
                                    matcher)
                                ,(if (symbolp body)
                                    `,(symbol-function body)
                                    body))
                           rule-cache)))
             (setf (symbol-function 'assert!)
                   #'(lambda (fact just)
                        ;; (format t "~%Assert! called")
                        (push `(assert! ,fact ,just) fact-cache)))
             (setf (symbol-function 'assume!)
                   #'(lambda (fact just)
                        ;; (format t "~%Assume! called")
                        (push `(assume! ,fact ,just) fact-cache)))
             ;;This eval *should* be OK bc it isn't called at runtime
             (mapc #'eval forms))
        (setf (symbol-function 'assume!) safe-assume!
              (symbol-function 'assert!) safe-assert!
              (symbol-function 'insert-rule) safe-insert-rule))
     (values rule-cache fact-cache)))


#+DEBUG-LTRE-RULES
(defun make-ltre-cache (&rest forms)
  ;; For each procedure that installs stuff
  ;; into an LTRE, redefine so that it instead
  ;; caches an appropriate calling form onto
  ;; one of rule-cache or fact-cache
  (let ((rule-cache nil)
        (fact-cache nil)
        (safe-insert-rule (symbol-function 'insert-rule))
        (safe-assert! (symbol-function 'assert!))
        (safe-assume! (symbol-function 'assume!)))
     ;; if an exit occurs in cleanup form we are in trouble
     (unwind-protect
          (progn
             (setf (symbol-function 'insert-rule)
                   ;;;Add var name after to lambda list after dbclass when debugging rules
                   #'(lambda (dbclass NAME matcher body)
                        ;; ACLPC is fussy about using symbols as procedures
                        ;; In fact, we have to place function objects in cache
                        ;; in order to avoid compile problems in runtime
                        (push `(,(dbclass-name dbclass)
                                ,NAME
                                ,(if (symbolp matcher)
                                    `,(symbol-function matcher)
                                    matcher)
                                ,(if (symbolp body)
                                    `,(symbol-function body)
                                    body))
                           rule-cache)))
             (setf (symbol-function 'assert!)
                   #'(lambda (fact just)
                        ;; (format t "~%Assert! called")
                        (push `(assert! ,fact ,just) fact-cache)))
             (setf (symbol-function 'assume!)
                   #'(lambda (fact just)
                        ;; (format t "~%Assume! called")
                        (push `(assume! ,fact ,just) fact-cache)))
             ;;This eval *should* be OK bc it isn't called at runtime
             (mapc #'eval forms))
        (setf (symbol-function 'assume!) safe-assume!
              (symbol-function 'assert!) safe-assert!
              (symbol-function 'insert-rule) safe-insert-rule))
     (values rule-cache fact-cache)))

;;; ----------------------------------------------------------------------------
;;; END OF CODE
