;;;; -*-  Mode: LISP; Syntax: Common-Lisp; Base: 10                          -*-
;;;; ---------------------------------------------------------------------------
;;;; File name: ltms.lsp
;;;;    System: LTRE
;;;;   Version: 4.1 
;;;;    Author: Ken Forbus, Johan de Kleer, John Everett
;;;;   Created: July 2, 1992
;;;;  Modified: everett on Fri Aug 2 9:18:03 1996
;;;;   Purpose: Logic-based truth maintenance system, version
;;;; Copyright: 1986, 1989, 1990, 1991, 1995 Kenneth D. Forbus, 
;;;;            Northwestern University, and Johan de Kleer, Xerox Corporation.
;;;;            All Rights Reserved.
;;;; ---------------------------------------------------------------------------

(in-package :LTRE)


;;; DEFINITIONS
;;; ----------------------------------------------------------------------------
(defvar *ltms* nil "Binding for current LTMS used in normalizing formulas")
                
(defstruct (ltms (:PRINT-FUNCTION print-ltms))
   (title nil)
   (node-counter 0)              ; unique namer for nodes.
   (clause-counter 0)            ; unique namer for justifications.
   (clause-mark 0)               ; Flip-flop mark map-over-clauses uses
   (nodes nil)                   ; hash table for nodes.
   (checking-contradictions t)
   (node-string nil)			       
   (contradiction-handlers nil)
   (pending-contradictions nil)
   (enqueue-procedure nil)
   (node-thread-dirty? nil)      ; Indicates ltms in midst of retraction
   (clause-thread-dirty? nil)    ; Indicates ltms in midst of set-truth
   (complete nil)                ; Is this a complete LTMS?
   (violated-clauses nil)
   ;;CLTMS SLOTS
   (queue nil)			 ; CLTMS queue of clauses to resolve.
   (conses nil)			 ; Source of conses to reuse in inner loop.
   (delay-sat nil)		 ; Don't resolve satisfied clauses.
   (cons-size 0))		 ; Size of temporary structure.

(defstruct (tms-node (:PRINT-FUNCTION print-tms-node))
   ;;Changes to this struct require recompilation of rule files
   (index 0)                    ; unique namer for nodes
   (datum nil)		        ; positive inference engine datum.
   (label :UNKNOWN)             ; :UNKNOWN, :TRUE, or :FALSE.
   (support nil)                ; clause which supports it
   (true-clauses nil)           ; clauses in which this node is true
   (false-clauses nil)          ; clauses in which this node is false
   (mark nil)                   ; marker for sweep algorithms
   (assumption? nil)            ; flag indicating whether or not node is an asn
   (true-rules nil)             ; rules run when the node is true
   (false-rules nil)            ; rules run when the node is false
   (true-rule-bindings nil)     ; cached of bindings for true rules
   (false-rule-bindings nil)    ; cached of bindings for false rules
   (rules nil)                  ; rules that matched on node
   (rules-to-gc nil)            ; rules that contain node in closures
   (ltms nil)                   ; LTMS it is part of.
   (true nil)		        ; True literal.
   (false nil)                  ; False literal. 
   (thread nil)                 ; pointer used in retractions
   )

(defstruct (clause (:PRINT-FUNCTION print-clause))
   (index 0)       ; Unique namer
   (informant nil) ; user-supplied tag for clause, helps in debugging
   (literals nil)  ; a list of (<node> . <truth>)
   (pvs 0)         ; Number of terms which potentially violate it.
   (length 0)      ; Number of literals.
   (sats 0)	   ; Number of terms which satisfy it.
   (status 0)      ; LTMS> 0/1 CLTMS> :SUBSUMED | :QUEUED | :DIRTY | :NOT-INDEXED | nil
   (thread nil)    ; pointer to minimize consing in set-truth
   (rule nil)      ; cache for reinstalling clause
   (rule-env nil)) ; cache for reinstalling clause

(defun print-ltms (ltms stream ignore)
   (declare (ignore ignore))
  (format stream "#<LTMS: ~A>" (ltms-title ltms)))

(defun clause-eq (c1 c2)
   (= (clause-index c1) (clause-index c2)))

(defun literal-eq (lit1 lit2)
  "A literal is a cons of a node and a belief condition"
   (and (node-eq (car lit1) (car lit2))
        (eq (cdr lit1) (cdr lit2))))

(defun node-eq (n1 n2)
   (= (tms-node-index n1) (tms-node-index n2)))

(defun print-tms-node (node stream ignore)
   (declare (ignore ignore))
   (if (tms-node-p node)
      (format stream "#<NODE ~A: ~A>" 
         (tms-node-index node) (node-string node))
      (format stream "<Null node: ~A>" node)))

(defun count-nodes ()
   "Does not rely on ltms-node-counter due to gc of nodes"
   (let ((count 0))
      (map-over-tms-nodes #'(lambda (n) (declare (ignore n)) (incf count)))
      count))

(defun count-clauses ()
   "Does not rely on ltms-clause-counter due to gc of nodes"
   (let ((count 0))
      (map-over-clauses #'(lambda (n) (declare (ignore n)) (incf count)))
      count))

(defun print-clause (clause stream ignore)
   (declare (ignore ignore))
   (format stream "#<Clause-~D ~A>"
      (clause-index clause) (clause-informant clause)))

(defun node-string (node)
   "Checks to see if arg is node prior to printing"
   (if (tms-node-p node)
      (funcall (ltms-node-string (tms-node-ltms node)) node)
      (format nil "~S is not a node" node)))

(define-condition node-label-error (simple-error)
   ((node :initarg :node :accessor node)
    (assumed-label :initarg :assumed-label :accessor assumed-label))
   (:report
    (lambda (condition stream)
       (with-slots (node assumed-label) condition
          (format stream 
             "~&LTMS error: You have tried to assume that ~A is ~A~%~
                but it already has a ~A label"
             (node-string node) assumed-label (tms-node-label node))))))
               
(defun node-label-error (node assumed-label)
   (error (make-condition 'node-label-error
             :node node
             :assumed-label assumed-label)))

(defun ltms-error (string &optional thing) (error string thing))

(defun default-node-string (n)
  (format nil "~A" (tms-node-datum n)))

(defmacro satisfied-clause? (clause) `(> (clause-sats ,clause) 0))
(defmacro violated-clause? (clause) `(= (clause-pvs ,clause) 0))

(defmacro walk-clauses (ltms f)
   `(if (ltms-complete ,ltms)
       (walk-trie ,f (ltms-clauses ,ltms))
       (map-over-clauses ,f)))


;;; BASIC INFERENCE-ENGINE INTERFACE
;;; ----------------------------------------------------------------------------
;;; Because garbage collection requires a tighter linking of ltre and ltms, 
;;; debugging instrumentation is now entirely in the hands of the LTRE, since that 
;;; is the interface presented to the user.

(defun create-ltms (title &key (node-string 'default-node-string)
 			       (checking-contradictions T)
			       (contradiction-handler 'ask-user-handler)
                               (cache-datums? T) ;; For stand-alone use
			       (complete nil)
			       (delay-sat T)
			  &aux ltms)
  (setq ltms
	(make-ltms :TITLE title
		   :NODES (if cache-datums? (make-hash-table :TEST #'eq))
		   :NODE-STRING node-string
		   :CHECKING-CONTRADICTIONS checking-contradictions
                   :CONTRADICTION-HANDLERS (list contradiction-handler)
                   :DELAY-SAT delay-sat
                   :COMPLETE complete))
   ltms)

(defun change-ltms (ltms &key (contradiction-handler nil contra?)
		              node-string
                              (checking-contradictions nil checking?)
                              (complete nil complete?)
			      (delay-sat nil delay-sat?))
   (if node-string (setf (ltms-node-string ltms) node-string))
   (if checking? (setf (ltms-checking-contradictions ltms)
                       checking-contradictions))
   (if contra?
      (setf (ltms-contradiction-handlers ltms)
            (list contradiction-handler)))
   (if complete? (setf (ltms-complete ltms) complete))
   (if delay-sat? (setf (ltms-delay-sat ltms) delay-sat)))


;;; NODE PREDICATES
;;; ----------------------------------------------------------------------------
(defun unknown-node? (node) (eq (tms-node-label node) :UNKNOWN))
(defun known-node?   (node) (not (eq (tms-node-label node) :UNKNOWN)))
(defun true-node?    (node) (eq (tms-node-label node) :TRUE))
(defun false-node?   (node) (eq (tms-node-label node) :FALSE))


;;; CREATING NODES
;;; ----------------------------------------------------------------------------
(defun tms-create-node (ltms datum &key assumptionp)
  (if (and (ltms-nodes ltms) (gethash datum (ltms-nodes ltms)))
      (ltms-error "Two nodes with same datum:" datum))
  (let ((node (make-tms-node :INDEX (incf (ltms-node-counter ltms))
			     :DATUM datum
			     :ASSUMPTION? assumptionp
			     :LTMS ltms)))
    (setf (tms-node-true node) (cons node :TRUE))
    (setf (tms-node-false node) (cons node :FALSE))
    (if (ltms-nodes ltms)
	(setf (gethash datum (ltms-nodes ltms)) node))
    (when (and (ltms-complete ltms)
	       (> (ltms-node-counter ltms) (ltms-cons-size ltms)))
      (setf (ltms-conses ltms) nil)
      (incf (ltms-cons-size ltms) 50.)
      (dotimes (i (ltms-cons-size ltms))
	(push (cons nil nil) (ltms-conses ltms))))
    node))

(defun enable-assumption (node label &optional (asn-type :ENABLED-ASSUMPTION))
   "ASN-TYPE must be a symbol"
   (cond ((not (tms-node-assumption? node))
          (ltms-error "Can't enable the non-assumption ~A" node))
         ((eq (tms-node-label node) label)
          (setf (tms-node-support node) asn-type))
         ((eq (tms-node-label node) :UNKNOWN)
          (top-set-truth node label asn-type))
         (t (node-label-error node label))))

(defun enabled-assumption? (node)
  (and (tms-node-support node)
       (not (typep (tms-node-support node) 'clause))))

(defun convert-to-assumption (node)
   (unless (tms-node-assumption? node)
      (dbg-ltre :LTMS "Converting ~A into an assumption" node)
      (setf (tms-node-assumption? node) T)))

(defun retract-assumption (node &aux ltms)
   (when (and (known-node? node)
              (enabled-assumption? node))
     (setq ltms (tms-node-ltms node))
     (setf (ltms-node-thread-dirty? ltms) T)
     (gc-node-thread-as-needed ltms
      (find-alternative-support ltms
       (propagate-unknownness node)))))


;;; COLLECTING FACT GARBAGE
;;; ----------------------------------------------------------------------------
;;; If we gc anything to do with a non-gcable node (ie, suppose the asn (foo bar) 
;;; generated a whole lot of nvalue statements in the database), then we lose 
;;; clause structure that will never get rebuilt.  Therefore this procedure 
;;; only gc's when the initial node is gc-able, hence the safe-to-gc? flag.
;;; See gc-preds.lsp for definition of gc-able-node?.

(defun gc-node-thread-as-needed (ltms node &aux next-node)
   "GC any nodes that can be and reset thread pointers"
   (do ((the-node node next-node))
       ((null the-node) t)
      (when (gc-able-node? the-node)
         (collect-node ltms the-node))
      ;;progress down the thread, resetting node pointers
      ;;Note: (shiftf a b c) ==> (setq a b) (setq b c)
      (shiftf next-node (tms-node-thread the-node) nil))
   (scour-rules) ;;trying this again
   (scour-nodes) ;;have to sweep for dead rules
   (setf (ltms-node-thread-dirty? ltms) nil))

(defun collect-node (ltms node &optional (create-thunks? t) &aux datum)
   "Setting create-thunks? to nil will cause permanent erasure of information"
   (setq datum (tms-node-datum node))
   (dbg-ltre :GC "~&GCing ~A~%" (datum-lisp-form datum))
   (delete-node-clauses node create-thunks?)
   (delete-bindings node (tms-node-rules node))
   (delete-rules (tms-node-rules-to-gc node))
   (delete-node ltms node datum)
   (delete-datum datum))

;;; Note: To avoid load-order dependencies, see runrules.lsp for definition of 
;;; delete-bindings, delete-rule and delete-node-from-rule-node-cache and 
;;; ldata.lsp for definition of delete-datum

(defun delete-node (ltms node datum)
   (setf (tms-node-label node) :GCD)
   (when (ltms-nodes ltms)
      (remhash datum (ltms-nodes ltms)))
   (setf (tms-node-datum node) nil)
   (dolist (rule (tms-node-rules node))
      (delete-node-from-rule-node-cache rule node))
   (setf (tms-node-rules node) nil)
   (setf (tms-node-rules-to-gc node) nil))


;;; DELETING CLAUSES
;;; ----------------------------------------------------------------------------
;;; Deleting clauses can in some cases cause a permanent loss of information
;;; from the ltms.  The solution to this problem is to allow gc to progress, but
;;; to create restoring thunks for those clauses that would otherwise be lost.
;;; The problem occurs when we have the following support relationship:
;;; gc-able -> non-gc-able -> gc-able.  In this case the gc-able nodes and their 
;;; clauses are deleted, but because the non-gc-able remains in place, no rule to 
;;; rebuild the lost clause(s) will execute.  In this situation we move some info
;;; that had been cached on the clause (see below and also bldrules.lsp) at the
;;; time of clause creation onto the innermost unknown and non-gc-able node's 
;;; caches, which the standard LTMS scheduling mechanisms will service when that 
;;; node's truth condition changes (ie, when another gc-able node that sets the 
;;; truth of the non-gc-able node is inserted into the database).

(defun delete-node-clauses (node &optional (create-thunks? t))
   (delete-clauses (tms-node-true-clauses node) create-thunks?)
   (delete-clauses (tms-node-false-clauses node) create-thunks?))

(defun delete-clauses (clauses create-thunks?)
   (dolist (clause clauses)
      (when create-thunks?
         (create-restoring-thunk clause))
      (setf (clause-rule clause) nil)
      (setf (clause-rule-env clause) nil)
      (dolist (lit (clause-literals clause))
         (excise-clause-from-literal (car lit) (cdr lit) clause))))

(defun excise-clause-from-literal (node label clause)
   (ecase label
      (:FALSE 
       (setf (tms-node-false-clauses node)
             (delete clause (tms-node-false-clauses node) :test #'eq)))
      (:TRUE
       (setf (tms-node-true-clauses node)
             (delete clause (tms-node-true-clauses node) :test #'eq)))))

;;; JE 3/22/96 Because LTRE users are free to call assert!/assume! from the 
;;; toplevel (eg, CyclePad does this in cprop when it accepts proposed 
;;; nvalues) it is quite possible that a clause will be created with no 
;;; current-rule/env.  In this case we cannot create a restoring thunk, else we 
;;; end up applying rules to nil and erroring out.

(defun create-restoring-thunk (clause &aux env rule literal)
   ;;; Note that this will only happen when there is an unknown and non-gc-able 
   ;;; trigger node in the antecedents of the clause.  Fortunately, this should
   ;;; be the only situation in which a clause will need to be reinstalled.
   (when (and (setq env (clause-rule-env clause))
              (setq rule (clause-rule clause))
              (setq literal (seek-literal clause)))
      (let ((label (cdr literal))
            (node (car literal)))
         (ecase label
            (:FALSE ;;these are DNF antecedent labels, so use opposite
             (unless (find rule (tms-node-true-rules node) :test #'rule-eq)
                (push rule (tms-node-true-rules node))
                (push env (tms-node-true-rule-bindings node))))
            (:TRUE ;;these are DNF antecedent labels, so use opposite
             (unless (find rule (tms-node-false-rules node) :test #'rule-eq)
                (push rule (tms-node-false-rules node))
                (push env (tms-node-false-rule-bindings node))))))))
   
(defun seek-literal (clause)
   (let ((trigger-sets (rule-triggers (clause-rule clause))))
      (car (member-if
              #'(lambda (n)
                   (and (unknown-node? n)
                        (some #'(lambda (trigger-set)
                                   (member n trigger-set))
                           trigger-sets)
                        (not (gc-able-node? n))))
              (clause-literals clause)
              :key #'car))))


;;; GC SWEEP FUNCTIONS
;;; ----------------------------------------------------------------------------
;;; The scour functions sweep through all nodes, deleting any rules that aren't 
;;; found in the dbclass table.  These stuck rules occur when an inner rule has 
;;; two or more enclosing rules that each are gc'able.  In this case it must
;;; install itself on the trigger node of each rule, since there is no way to
;;; know which node will be gc'd.  When one node gets gc'd, the other nodes 
;;; must be relieved of their rules.  Note that this sweep could be avoided by 
;;; establishing pointers amongst all the nodes containing the same rules, but
;;; that would be complicated and probably not worth the incremental speedup.

(defun scour-nodes (&aux rules)
   "Remove any rules not in dbclass table from node caches"
   (map-over-tms-nodes
      #'(lambda (n)
           (when (setq rules (tms-node-rules-to-gc n))
              (setf (tms-node-rules-to-gc n) 
                    (delete-if-not 'in-dbclass? rules)))
           (when (setq rules (tms-node-rules n))
              (setf (tms-node-rules n)
                    (delete-if-not 'in-dbclass? rules))))))

(defun in-dbclass? (rule)
   "Rules not in dbclass have been gc'd"
   (member rule (dbclass-rules (rule-dbclass rule)) :test #'eq))

(defparameter *scour-rules-counter* 0)

(defun scour-rules ()
   (map-over-rules 
      #'(lambda (r)
           (unless (in-dbclass? r)
              (incf *scour-rules-counter*)
              (setf (rule-tms-nodes r) nil)))))
   

;;; ADDING FORMULAS TO THE LTMS.
;;; ----------------------------------------------------------------------------
(defun add-formula (ltms formula &optional informant)
  (setq informant (list :IMPLIED-BY formula informant))
  (dolist (clause (normalize ltms formula))
    (unless (eq :TRUE (setq clause (simplify-clause clause)))
	(add-clause-internal clause informant T)))
  (check-for-contradictions ltms))

(defun simplify-clause (literals)
  (setq literals (sort-clause literals))
  (do ((tail literals next)
       (next (cdr literals) (cdr next)))
      ((null next) literals)
    (cond ((not (eq (caar tail) (caar next))))
	  ((not (eq (cdar tail) (cdar next)))
	   (return-from simplify-clause :TRUE))
	  (t (rplacd tail (cdr next))))))

(defun sort-clause (literals)
   (sort (copy-list literals) #'< 
      :KEY #'(lambda (n) (tms-node-index (car n)))))

(defun normalize (*ltms* exp) (normalize-1 exp nil))

(defun normalize-1 (exp negate)
  (case (and (listp exp) (car exp))
    (:IMPLIES (if negate
		  (nconc (normalize-1 (cadr exp) nil)
			 (normalize-1 (caddr exp) t))
		  (disjoin (normalize-1 (cadr exp) t)
			   (normalize-1 (caddr exp) nil))))
    (:IFF (normalize-iff exp negate))
    (:OR (if negate (normalize-conjunction exp t)
	     (normalize-disjunction exp nil)))
    (:AND (if negate (normalize-disjunction exp t)
	             (normalize-conjunction exp nil)))
    (:NOT (normalize-1 (cadr exp) (not negate)))
    (:TAXONOMY (normalize-tax exp negate))
    (t (if negate `((,(tms-node-false (find-node *ltms* exp))))
	          `((,(tms-node-true (find-node *ltms* exp))))))))

(defun normalize-tax (exp negate)
  (normalize-1 `(:AND (:OR ,@ (copy-list (cdr exp))) ;one must be true
	       ,@ (do ((firsts (cdr exp) (cdr firsts))
		       (rests (cddr exp) (cdr rests))
		       (result nil))
		      ((null rests) result)
		    (dolist (other rests)
		      (push `(:NOT (:AND ,(car firsts) ,other))
			    result))))
	       negate))

(defun normalize-conjunction (exp negate)
  (mapcan #'(lambda (sub) (normalize-1 sub negate)) (cdr exp)))

(defun normalize-iff (exp negate)
  (nconc (normalize-1 `(:IMPLIES ,(cadr exp) ,(caddr exp)) negate)
	 (normalize-1 `(:IMPLIES ,(caddr exp) ,(cadr exp)) negate)))

(defun normalize-disjunction (exp negate)
  (unless (cdr exp)
    (return-from normalize-disjunction (list nil)))
  (do ((result (normalize-1 (cadr exp) negate))
       (rest (cddr exp) (cdr rest)))
      ((null rest) result)
    (setq result (disjoin (normalize-1 (car rest) negate) result))))

(defun disjoin (conj1 conj2)
  (unless (or conj1 conj2) (return-from disjoin nil))
  (mapcan #'(lambda (disj1)
	    (mapcar #'(lambda (disj2) (append disj1 disj2))
		    conj2))
	  conj1))

(defun find-node (ltms name)
  (cond ((typep name 'tms-node) name)
	((if (ltms-nodes ltms) (gethash name (ltms-nodes ltms))))
	((tms-create-node ltms name))))

(defmacro compile-formula (run-tms f &optional informant &aux ltms)
  (setq ltms (create-ltms f))
  (add-formula ltms (expand-formula f))
  (generate-code ltms run-tms (if informant `(:IMPLIED-BY ,f ,informant))))

(defun generate-code (ltms run-tms informant &aux result bound datum)
  (maphash #'(lambda (ignore symbol)
	       (declare (ignore ignore))
	       (when (or (tms-node-true-clauses symbol)
			 (tms-node-false-clauses symbol))
		 (setq datum (tms-node-datum symbol))
		 (when (listp datum)
		   (setf (tms-node-mark symbol) datum)
		   (setf (tms-node-datum symbol)
			 (make-symbol (format nil "~A" (cadr datum))))
		   (push symbol bound))))
	   (ltms-nodes ltms))
  (walk-clauses ltms
		#'(lambda (clause &aux ps ns)
		    (dolist (lit (clause-literals clause))
		      (if (eq (cdr lit) :TRUE)
			  (push (tms-node-datum (car lit)) ps)
			  (push (tms-node-datum (car lit)) ns)))
		    (push `(add-clause `(,,@ps) `(,,@ns) ,informant)
			  result)))
  `(let ,(mapcar #'(lambda (s)
		     `(,(tms-node-datum s) (find-node ,run-tms ,(tms-node-mark s))))
		 bound)
     ,@result))

(defun expand-formula (x)
  (setq x (macroexpand x))
  (cond ((not (listp x)) x)
	((case (macroexpand (car x))
	   (QUOTE (partial (cadr x)))
	   (LIST (mapcar #'expand-formula (cdr x)))
	   (LIST* (if (cddr x)
		      (cons (expand-formula (cadr x))
			    (expand-formula `(LIST* .,(cddr x))))
		      (expand-formula (cadr x))))
	   (CONS (cons (expand-formula (cadr x))
		       (mapcar #'expand-formula (caddr x))))))
	(t x)))

(defun partial (x)
  (cond ((null x) x)
	((keywordp x) x)
	((not (listp x)) `',x)
	(t (cons (partial (car x)) (partial (cdr x))))))


;;; ADDING CLAUSES
;;; ----------------------------------------------------------------------------
;;; Clauses are made in bcp-add-clause, and three pieces of information are 
;;; cached on each clause, the current rule, a list of non-gc-able triggers and
;;; the environment the current rule will need when run to reinstate the clause.
;;; The dynamic var *non-gc-able-triggers* is bound only when there are both 
;;; gc-able and non-gc-able triggers in the antecedents of the clause, since this 
;;; is the only situation in which we want to cache information on the clause.  
;;; This var is lambda-bound within rule bodies that create clauses; see 
;;; bldrules.lsp for this code.

;;; Generalizing the above: Testing an alternative in which we always cache the 
;;; current rule and its current env on the clause, which will increase clause 
;;; bloat.  However, the theory is that this will enable us to gc facts that 
;;; are the consequents of non-gc-able nodes.

(defun add-clause (true-nodes false-nodes &optional informant)
  (add-clause-internal (nconc (mapcar #'tms-node-true true-nodes)
			      (mapcar #'tms-node-false false-nodes))
		       informant
		       nil))

(defun add-clause-internal (literals informant internal &aux ltms)
   (setq ltms (tms-node-ltms
                 (or (caar literals)
                     (ltms-error "Total contradiction: Null clause" informant))))
   (if (ltms-complete ltms)
      (full-add-clause ltms literals informant)
      (bcp-add-clause ltms literals informant))
   (unless internal (check-for-contradictions ltms)))

;;; Because LTRE users are free to call assert!/assume! from the 
;;; toplevel (eg, CyclePad does this in cprop when it accepts proposed 
;;; nvalues) it is quite possible that a clause will be created with no 
;;; current-rule/env.

(defun bcp-add-clause (ltms literals informant &optional (index T)
                         &aux cl label)
   (setq cl (make-clause 
               :INDEX (incf (ltms-clause-counter ltms))
               :LITERALS literals
               :INFORMANT informant
               :LENGTH (length literals)
               :RULE *current-rule*
               :RULE-ENV *current-env*))
   (dolist (term literals)
      (when (eq :UNKNOWN (setq label (tms-node-label (car term))))
         (incf (clause-pvs cl)))
      (ecase (cdr term)
         (:TRUE
          (when index (insert-true-clause cl (car term)))
          (when (eq label :TRUE)
             (incf (clause-sats cl)) (incf (clause-pvs cl))))
         (:FALSE
          (when index (insert-false-clause cl (car term)))
          (when (eq label :FALSE)
             (incf (clause-sats cl)) (incf (clause-pvs cl))))))
   (when index (check-clauses ltms cl))
   (dbg-ltre :clauses "BCP-add-clause returning ~A" cl)
   cl)

(defun insert-true-clause (cl node)
  (push cl (tms-node-true-clauses node)))

(defun insert-false-clause (cl node)
  (push cl (tms-node-false-clauses node)))

(defun add-nogood (culprit sign assumptions &aux trues falses)
   (dolist (a assumptions (add-clause trues falses 'NOGOOD))
      (ecase (if (eq a culprit) sign (tms-node-label a))
         (:TRUE (push a falses))
         (:FALSE (push a trues)))))

;;; ----------------------------------------------------------------------------
;;; END OF CODE